#! /usr/bin/env python
"""Command-Line Interface for read_acq."""

import click
import glob
import tqdm

from . import convert_file

main = click.Group()


@main.command()
@click.argument(
    "infile",
    nargs=-1,
)
@click.option(
    "-o",
    "--outfile",
    default=None,
    type=click.Path(exists=False, dir_okay=False),
)
@click.option(
    "-f",
    "--fmt",
    default="h5",
    type=click.Choice(["h5", "mat", "npz"]),
)
def convert(infile, outfile, fmt):
    """Convert an ACQ file to a different format."""
    fls = []
    for fl in infile:
        fls += glob.glob(fl)
    fls = list(set(fls))

    for fl in tqdm.tqdm(
        fls, disable=len(fls) < 5, desc="Processing files", unit="files"
    ):
        convert_file(fl, outfile=outfile, write_format=fmt)



import pytest

import numpy as np
from pathlib import Path

from read_acq import convert_file, convert_h5, decode_file, encode
from read_acq.codec import _decode_line, _encode, _encode_line
from read_acq.read_acq import (
    ACQError,
    ACQLineError,
    Ancillary,
    CommentLine,
    DataEntry,
    DataLine,
)


def test_roundtrip():
    a = np.logspace(-5, -1, 20)

    b = _decode_line(_encode(a))
    assert np.allclose(a, b)


def test_roundtrip_ratio():
    a = np.logspace(-5, -1, 20)
    b = np.logspace(-6, -2, 20)

    aa = _decode_line(_encode_line(a, 1000))
    bb = _decode_line(_encode_line(b, 1000))

    assert np.allclose(a[10:] / b[10:], aa[10:] / bb[10:])


@pytest.fixture(scope="module")
def sample_fastspec_acq():
    return Path(__file__).parent / "data/sample.acq"


@pytest.fixture(scope="module")
def sample_pxspec_acq():
    """First six lines from a real file.

    /data5/edges/data/2014_February_Boolardy/mro/low/2016/2016_250_02.acq
    """
    return Path(__file__).parent / "data/pxspec.acq"


@pytest.fixture(scope="module")
def sample_acq_file(tmp_path_factory):
    data = np.logspace(-5, -1, 1500).reshape((3, 5, 100))
    meta = {
        "temp": 0,
        "nblk": 1000,
        "nfreq": 100,
        "freq_min": 0,
        "freq_max": 200,
        "freq_res": 2,
    }

    fname = tmp_path_factory.mktemp("direc") / "tempfile.acq"

    ancillary = {
        "adcmin": np.ones((10, 3)) * 0.2,
        "adcmax": np.ones((10, 3)) * 0.3,
        "data_drops": np.zeros((10, 3), dtype=int),
        "times": np.array(["2016:080:01:01:01"] * 10),
    }

    encode(fname, data, meta, ancillary)
    return fname


@pytest.fixture(scope="module")
def incomplete_file(sample_acq_file):
    with open(sample_acq_file) as fl:
        lines = fl.readlines()
    last_line = lines[-1]
    meta, spec = last_line.split(" spectrum ")
    spec = _decode_line(spec.lstrip())
    spec = spec[: spec.size // 2]
    spec = _encode_line(spec, nblk=1000)
    lines[-1] = meta + " spectrum " + spec + "\n"

    new_file = sample_acq_file.parent / "incomplete.acq"
    with open(new_file, "w") as fl:
        fl.writelines(lines)
    return new_file


def test_full_file_encode(sample_acq_file):
    with open(sample_acq_file) as fl:
        assert fl.readline() == ";--temp: 0\n"


def test_full_file_h5(sample_acq_file, tmp_path):
    h5_file = tmp_path / "tempfile.h5"
    convert_file(sample_acq_file, outfile=h5_file)

    new_acq = tmp_path / "tempfile_roundtrip.acq"
    # Now convert back to .acq
    convert_h5(h5_file, new_acq)

    with open(new_acq) as fl:
        assert ";--temp: 0\n" in fl.readlines()[:20]


def test_incomplete_file(incomplete_file):
    Q, p = decode_file(incomplete_file, progress=False)

    # We lost an integration!
    for pp in p:
        assert pp.T.shape == (4, 100)
    assert Q.shape == (4, 100)


@pytest.fixture(scope="module")
def pxspec_comment_line():
    """A real comment-line from a pxspec-output file.

    Taken from /data5/edges/data/2014_February_Boolardy/mro/low/2016/2016_250_02.acq
    """
    return (
        "# swpos 0 resolution   24.414 adcmax  0.28123 adcmin -0.31032 temp  0 C "
        "nblk 40960 nspec 32768"
    )


@pytest.fixture(scope="module")
def data_line_prefix():
    """A real data-line prefix from a pxspec-output file.

    Doesn't include the actual spectrum here.
    Taken from /data5/edges/data/2014_February_Boolardy/mro/low/2016/2016_250_02.acq
    """
    return (
        "2016:250:02:06:07 0    0.000 0.006104  200.000  0.3 spectrum /fabricated-data/"
    )


@pytest.fixture(scope="module")
def fastspec_comment_line():
    """A real comment-line from a fastspec-output file.

    From 2014_February_Boolardy/mro/low2/2022/2022_010_16_12_17_low2.acq,
    but manually changed swpos to 1 to enable some "broken" tests below.
    """
    return (
        "# swpos 1 data_drops        0 adcmax  0.21106 adcmin -0.22217 temp  0 C "
        "nblk 65532 nspec 32768"
    )


def test_pxspec_comment_line(pxspec_comment_line):
    c = CommentLine.read(pxspec_comment_line)
    assert c.swpos == 0
    assert c.resolution == 24.414
    assert c.adcmax == 0.28123
    assert c.adcmin == -0.31032
    assert c.temp == 0
    assert c.nblk == 40960
    assert c.nspec == 32768
    assert c.data_drops is None


def test_fastspec_comment_line(fastspec_comment_line):
    c = CommentLine.read(fastspec_comment_line)
    assert c.swpos == 1
    assert c.data_drops == 0
    assert c.adcmax == 0.21106
    assert c.adcmin == -0.22217
    assert c.temp == 0
    assert c.nblk == 65532
    assert c.nspec == 32768
    assert c.resolution is None

    with pytest.raises(ACQError, match="Could not parse line"):
        CommentLine.read(fastspec_comment_line, fastspec=False)


def test_bad_comment_line():
    with pytest.raises(ACQError, match="Could not parse line"):
        CommentLine.read("this is not a comment line")


def test_data_line_prefix_read(data_line_prefix):
    d = DataLine.read(data_line_prefix, read_spectrum=False)
    assert d.time == "2016:250:02:06:07"
    assert d.swpos == 0
    assert d.freqmin == 0
    assert d.freqmax == 200
    assert d.deltaf == 0.006104
    assert d.thing == 0.3


def test_bad_data_line():
    with pytest.raises(ACQError, match="Could not parse line"):
        DataLine.read("this is not a data line spectrum  ", read_spectrum=False)


def test_read_data_entry(pxspec_comment_line, data_line_prefix):
    data_entry = DataEntry.read(
        (pxspec_comment_line, data_line_prefix), read_spectrum=False
    )
    assert data_entry.comment.swpos == 0
    assert data_entry.data.time == "2016:250:02:06:07"


def test_read_data_entry_mismatching_swpos(fastspec_comment_line, data_line_prefix):
    with pytest.raises(ACQLineError, match="swpos of comment and data do not match"):
        DataEntry.read((fastspec_comment_line, data_line_prefix), read_spectrum=False)


def test_read_data_entry_wrong_nspec(sample_fastspec_acq):
    # Get the two data-entry lines from the front of the file.
    lines = []
    with open(sample_fastspec_acq) as fl:
        for line in fl:
            if line.startswith("# "):
                lines.append(line)
                lines.append(next(fl))
                break

    lines[0] = lines[0].replace("nspec 32768", "nspec 32767")
    with pytest.raises(ACQLineError, match="nspec and length of spectrum do not match"):
        DataEntry.read(lines)


def test_ancillary_pxspec(sample_pxspec_acq):
    anc = Ancillary(sample_pxspec_acq)
    assert anc.fastspec_version is None
    assert "data_drops" not in anc.meta
    assert anc.meta["resolution"] == 24.414


def test_ancillary_append_bad(sample_pxspec_acq):
    anc = Ancillary(sample_pxspec_acq)

    # Read first two lines of data
    lines = []
    with open(sample_pxspec_acq) as fl:
        lines.append(fl.readline())
        lines.append(fl.readline())

    data_entry = DataEntry.read(lines)

    with pytest.raises(ValueError, match="swpos of datas must be"):
        anc.append((data_entry,) * 3)


def test_read_file_starting_with_swpos_1(sample_pxspec_acq, tmp_path):
    with open(sample_pxspec_acq) as fl:
        lines = fl.readlines()

    # Put the swpos=1 lines at the start of the file.
    lines = lines[2:4] + lines

    new = tmp_path / "order_test.acq"
    with open(new, "w") as fl:
        fl.writelines(lines)

    q, p = decode_file(new)
    # we should have just one complete switch cycle
    assert len(q) == 1
    assert p[0].shape[1] == 1
    assert p[1].shape[1] == 1
    assert p[2].shape[1] == 1


def test_read_file_with_incomplete_cycle(sample_pxspec_acq, tmp_path):
    with open(sample_pxspec_acq) as fl:
        lines = fl.readlines()

    # Make an imcomplete cycle in the middle.
    lines = lines + lines[:4] + lines

    new = tmp_path / "incomplete_test.acq"
    with open(new, "w") as fl:
        fl.writelines(lines)

    q, p = decode_file(new)
    # we should have just two complete switch cycles
    assert len(q) == 2
    assert p[0].shape[1] == 2
    assert p[1].shape[1] == 2
    assert p[2].shape[1] == 2


def test_read_file_with_incomplete_cycle_starting_bad(sample_pxspec_acq, tmp_path):
    with open(sample_pxspec_acq) as fl:
        lines = fl.readlines()

    # Make an imcomplete cycle in the middle. Now the first bad swpos is not zero.
    lines = lines + lines[2:6] + lines

    new = tmp_path / "incomplete_test2.acq"
    with open(new, "w") as fl:
        fl.writelines(lines)

    q, p = decode_file(new)
    # we should have just two complete switch cycles
    assert len(q) == 2
    assert p[0].shape[1] == 2
    assert p[1].shape[1] == 2
    assert p[2].shape[1] == 2



"""Package providing functionality for reading .ACQ format files."""
from pkg_resources import DistributionNotFound, get_distribution

from . import codec, writers
from .read_acq import Ancillary, convert_file, convert_h5, decode_file, encode

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    pass

del get_distribution
del DistributionNotFound



"""Functions that write ACQ file data to different formats."""
import copy
import numpy as np
from pathlib import Path
from scipy import io
from typing import Optional, Union

_WRITERS = []


def writer(func):
    """Decorator to denote a function as a writer."""
    fmt = func.__name__.split("_")[-1]
    _WRITERS.append(fmt)

    def writer_wrapper(outfile=None, ancillary=None, **data):
        outfile = _get_fname(outfile, fmt)
        print(f"Writing {outfile}...", end="", flush=True)
        func(outfile, ancillary, **data)
        print(" done")

    return writer_wrapper


def _get_fname(outfile: Optional[Union[Path, str]] = None, fmt: Optional[str] = None):
    if not outfile:
        raise Exception("You need to provide either outfile or infile!")
    outfile = Path(outfile)
    return outfile.with_suffix(f".{fmt}")


@writer
def _write_h5(outfile=None, ancillary=None, **data):
    """Writes HDF5 file exactly as FastSpec should now write it out."""
    try:
        from edges_io.h5 import HDF5RawSpectrum
    except ImportError:  # pragma: no cover
        raise ImportError(
            "To write to h5, you need to install edges_io or do "
            "`pip install read_acq[h5]"
        )

    spectra = {
        "Q": data["Qratio"],
        "p0": data["p0"],
        "p1": data["p1"],
        "p2": data["p2"],
    }
    ancillary.update(fastspec_version=data["fastspec_version"])
    freq_anc = {"frequencies": data["freqs"]}
    time_anc = {name: data["time_data"][name] for name in data["time_data"]}

    obj = HDF5RawSpectrum.from_data(
        {
            "spectra": spectra,
            "meta": ancillary,
            "freq_ancillary": freq_anc,
            "time_ancillary": time_anc,
        }
    )

    obj.write(outfile)


@writer
def _write_npz(outfile=None, ancillary=None, **data):
    data = copy.deepcopy(data)
    data.update(ancillary)
    np.savez(outfile, **data)


@writer
def _write_mat(outfile=None, ancillary=None, **data):
    data = copy.deepcopy(data)
    data.update(ancillary)
    io.savemat(outfile, data)



import h5py
import numpy as np
from click.testing import CliRunner
from pathlib import Path

from read_acq import decode_file
from read_acq.cli import convert


def test_convert(tmp_path_factory):
    cli = CliRunner()

    data = Path(__file__).parent / "data/sample.acq"
    outfile = tmp_path_factory.mktemp("direc") / "tempfile.h5"

    result = cli.invoke(convert, [str(data), "--outfile", str(outfile), "--fmt", "h5"])
    assert result.exit_code == 0
    print(result.output)

    Q, p, meta = decode_file(data, meta=True)

    with h5py.File(outfile, "r") as fl:
        qq = fl["spectra"]["Q"][...]

    assert np.allclose(qq[~np.isnan(qq)], Q[~np.isnan(Q)])



"""Wrapper for C-code that does the encoding-decoding."""
import ctypes
import glob
import numpy as np
import os

cdll = glob.glob(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), "decode.*.so")
)[0]
cdll = ctypes.CDLL(cdll)

_c_decode = cdll.decode
_c_decode.restype = ctypes.c_int
_c_decode.argtypes = [
    ctypes.c_char_p,
    np.ctypeslib.ndpointer(np.float64),
]

_c_encode = cdll.encode
_c_encode.restype = ctypes.c_int
_c_encode.argtypes = [
    ctypes.c_int,
    np.ctypeslib.ndpointer(np.float64),
    ctypes.c_char_p,
]


def _decode_line(line):
    """
    Decode a pre-parsed line of an ACQ file.

    This is a very thin wrapper around the C `decode` function which does the same.

    Parameters
    ----------
    line : str
        A string that is uencoded. There should be no spaces at the start or end of the
        line. This is *not* verified in this function.

    Returns
    -------
    np.ndarray :
        A 1D array of float values given by the de-encoding and unpacking. These are
        arbitrarily scaled linear powers.
    """
    out = np.zeros(len(line) // 4)
    res = _c_decode(ctypes.c_char_p(line.encode("ascii")), np.ascontiguousarray(out))

    if res:
        raise Exception("C decoder exited with an error!")

    return out


def _encode(data):
    """Encode an array of data.

    This is a low-level function that takes an array of linear-scaled float data and
    converts it to a string of 64-bit encoded integers. It does *not* perform scaling
    to match the output of fastspec, but its output should be the inverse of
    :func:`_decode_line`.
    """
    out = ctypes.create_string_buffer(len(data) * 4)
    res = _c_encode(len(data), data, out)

    if res:
        raise Exception("C encoder exited with an error!")

    return out.raw.decode("ascii")


def _encode_line(data, nblk):
    """Encode an array of data.

    This takes an array of linear-scaled float data and converts it to a string of
    64-bit encoded integers. It performs scaling to match the output of fastspec, and
    also blanks out the first 10 entries of data (which are never used in fastspec).

    The output is *not* exactly the inverse of :func:`_decode_line`, but is the same up
    to a scaling constant, as well as some clipping on dynamic range. Thus, while it is
    important that this function scales its input (to get the data into the clipping
    range), it is *not* important to scale the output of `_decode_line`, as it is
    fully linear, and only the ratio is ever used.
    """
    # We scale the data to achieve same dynamic range as in fastspec
    d = data / (2 * nblk * len(data) * 10**3.84)

    # Set first 10 values to 0 because they are never used
    d[:10] = 0

    return _encode(d)



#!/usr/bin/env python
"""Install the package."""

from setuptools import setup

from numpy.distutils.core import Extension

setup(
    use_scm_version=True,
    ext_modules=[
        Extension(
            "read_acq.decode",
            ["src/read_acq/decode.c"],
            extra_compile_args=["-Ofast", "-Wall"],
            libraries=["m"],
        ),
    ],
)



"""Functions and classes for reading and writing the .acq format."""
from __future__ import annotations

import attrs
import numpy as np
import re
import tqdm
import warnings
from pathlib import Path

from . import writers
from .codec import _decode_line, _encode_line


class ACQError(Exception):
    """Base class for errors in the ACQ module."""


class ACQLineError(ACQError):
    """Error in parsing a line of an ACQ file."""


@attrs.define()
class CommentLine:
    """A class for storing a comment line."""

    swpos: int = attrs.field(converter=int)
    adcmax: float = attrs.field(converter=float)
    adcmin: float = attrs.field(converter=float)
    temp: float = attrs.field(converter=float)
    nblk: int = attrs.field(converter=int)
    nspec: int = attrs.field(converter=int)
    resolution: float = attrs.field(
        converter=attrs.converters.optional(float), default=None
    )
    data_drops: int = attrs.field(
        converter=attrs.converters.optional(int), default=None
    )

    _sw = r"swpos (?P<swpos>\d)"
    _res = r"resolution \s*(?P<resolution>\d+(\.\d*)?|\.\d+)"
    _acmax = r"adcmax \s*(?P<adcmax>[-+]?\d+(\.\d*)?|\.\d+)"
    _acmin = r"adcmin \s*(?P<adcmin>[-+]?\d+(\.\d*)?|\.\d+)"
    _temp = r"temp \s*(?P<temp>\d+) C"
    _nblk = r"nblk \s*(?P<nblk>\d+)"
    _nspec = r"nspec \s*(?P<nspec>\d+)"
    _dd = r"data_drops \s*(?P<data_drops>\d+)"

    pxspec = re.compile(f"# {_sw} {_res} {_acmax} {_acmin} {_temp} {_nblk} {_nspec}")
    fastspec = re.compile(f"# {_sw} {_dd} {_acmax} {_acmin} {_temp} {_nblk} {_nspec}")

    @classmethod
    def read(cls, line, fastspec: bool | None = None):
        """Read an ACQ comment line as a CommentLine object."""
        line = line.strip()
        if fastspec is False:
            match = re.match(cls.pxspec, line)
        elif fastspec is True:
            match = re.match(cls.fastspec, line)
        else:
            match = re.match(cls.pxspec, line)
            if match is None:
                match = re.match(cls.fastspec, line)
        if match is None:
            raise ACQError(f"Could not parse line: '{line}'")

        return cls(**match.groupdict())


@attrs.define()
class DataLine:
    """A class for storing a data line."""

    time: str = attrs.field()
    swpos: int = attrs.field(converter=int)
    freqmin: float = attrs.field(converter=float)
    deltaf: float = attrs.field(converter=float)
    freqmax: float = attrs.field(converter=float)
    thing: float = attrs.field(converter=float)
    spectrum: np.ndarray | None = attrs.field(default=None)

    _time = r"(?P<time>\d{4}:\d{3}:\d{2}:\d{2}:\d{2})"
    _swpos = r"(?P<swpos>\d{1})"
    _float = r"[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?"
    _fmin = f"(?P<freqmin>{_float})"
    _df = f"(?P<deltaf>{_float})"
    _fmax = f"(?P<freqmax>{_float})"
    _last = f"(?P<thing>{_float})"

    regex = re.compile(rf"{_time} {_swpos} \s*{_fmin} \s*{_df} \s*{_fmax} \s*{_last}")

    @classmethod
    def read(cls, line, read_spectrum: bool = True):
        """Read an ACQ data line as a DataLine object."""
        front, back = line.split(" spectrum ")
        match = re.match(cls.regex, front)
        if match is None:
            raise ACQError(f"Could not parse line front-matter: {front}")

        if read_spectrum:
            spec = _decode_line(back.lstrip())
        else:
            spec = None

        return cls(spectrum=spec, **match.groupdict())


@attrs.define
class DataEntry:
    """A class for storing a data entry."""

    comment: CommentLine = attrs.field()
    data: DataLine = attrs.field()

    @data.validator
    def _check_data(self, attribute, value):
        if value.swpos != self.comment.swpos:
            raise ACQLineError("swpos of comment and data do not match")
        if value.spectrum is not None and value.spectrum.shape[0] != self.comment.nspec:
            raise ACQLineError("nspec and length of spectrum do not match")

    @classmethod
    def read(cls, lines, read_spectrum: bool = True) -> DataEntry:
        """Read an ACQ data entry (two lines) as a DataEntry object."""
        comment = CommentLine.read(lines[0])
        data = DataLine.read(lines[1], read_spectrum=read_spectrum)
        return cls(comment, data)


class Ancillary:
    """The ancillary data of an ACQ file."""

    header_char = ";"

    _splits = re.compile(r"[\d\.:]+")
    DTYPES = {
        "adcmax": np.float32,
        "adcmin": np.float32,
        "times": "S17",
    }

    def __init__(self, fname):
        self.fastspec_version = self._get_fastspec_version(fname)
        self.meta = self.read_metadata(fname)

        self.data = {
            "adcmax": [],  # np.zeros((self.size, 3), dtype=np.float32),
            "adcmin": [],  # np.zeros((self.size, 3), dtype=np.float32),
            "times": [],  # np.zeros((self.size, 3), dtype="S17"),
        }
        if "data_drops" in self.meta:
            self.data["data_drops"] = []  # np.zeros((self.size, 3), dtype=int)

    def _get_fastspec_version(self, fname):
        with open(fname) as fl:
            first_line = fl.readline()
        if first_line.startswith(self.header_char):
            return first_line.split("FASTSPEC")[-1]
        else:
            return None

    def _read_header(self, fname):
        out = {}

        name_pattern = re.compile(r"[a-zA-Z_]+")
        with open(fname) as fl:
            type_order = [int, float, str]

            for line in fl:
                if not line.startswith(self.header_char):
                    break

                if line.startswith("; FASTSPEC"):
                    name = "fastspec_version"
                    val = line.split()[-1]
                else:
                    try:
                        name, val = line.split(": ")
                    except ValueError:
                        warnings.warn(f"In file {fname}, item {line} has no value")
                        name = line.split(":")[0]
                        val = ""

                    name = name_pattern.findall(name)[0]

                for tp in type_order:
                    try:
                        out[name] = tp(val.split()[0])
                        break
                    except IndexError:
                        try:
                            out[name] = tp(val)
                        except ValueError:
                            pass
                    except ValueError:
                        pass

        return out

    def read_metadata(self, fname):
        """Read the metadata of the ACQ file."""
        out = self._read_header(fname)

        with open(fname) as fl:
            for line in fl:
                if line.startswith("#"):
                    comment = CommentLine.read(line)
                    data = DataLine.read(next(fl), read_spectrum=False)
                    break

        out.update(
            {
                "temperature": comment.temp,
                "nblk": comment.nblk,
                "nfreq": comment.nspec,
                "freq_min": data.freqmin,
                "freq_max": data.freqmax,
                "freq_res": data.deltaf,
            }
        )
        if comment.resolution is not None:
            out["resolution"] = comment.resolution
        if comment.data_drops is not None:
            out["data_drops"] = comment.data_drops

        return out

    @property
    def frequencies(self):
        """The frequencies associated with the spectrum measurements."""
        df = self.meta["freq_max"] / self.meta["nfreq"]

        # See edges-cal.tools.EdgesFrequencyRange for justification of using this
        # form.
        return np.arange(self.meta["freq_min"], self.meta["freq_max"], df)

    def append(self, datas: tuple[DataEntry, DataEntry, DataEntry]):
        """Append a set of data to the ancillary data."""
        if tuple(d.data.swpos for d in datas) != (0, 1, 2):
            raise ValueError(
                "swpos of datas must be (0, 1, 2), got "
                f"{tuple(d.data.swpos for d in datas)}"
            )

        self.data["adcmax"].append([d.comment.adcmax for d in datas])
        self.data["adcmin"].append([d.comment.adcmin for d in datas])
        self.data["times"].append([d.data.time for d in datas])

        if "data_drops" in self.meta:
            self.data["data_drops"].append([d.comment.data_drops for d in datas])

    def complete(self):
        """Convert the ancillary data to numpy arrays."""
        for key in self.data:
            self.data[key] = np.array(
                self.data[key], dtype=self.DTYPES.get(key, np.float32)
            )


def decode_file(
    fname,
    progress=True,
    meta=False,
    leave_progress=True,
):
    """
    Parse and decode an ACQ file, optionally writing it to a new format.

    Parameters
    ----------
    fname : str or Path
        filename of the ACQ file to read.

    progress: bool, optional
        Whether to display a progress bar for the read.
    meta: bool, optional
        Whether to output metadata for the read.
    leave_progress : bool, optional
        Whether to leave the progress bar (if one is used) on the screen when done.
        Useful to set to False if reading multiple files.
    """
    anc = Ancillary(fname)

    fastspec = "data_drops" in anc.meta

    p0, p1, p2 = [], [], []

    with open(fname) as fl:
        # First find the first swpos=0 line.

        for line in fl:
            if line.startswith("#"):
                cline = CommentLine.read(line, fastspec=fastspec)
                if cline.swpos == 0:
                    break

        data = DataLine.read(next(fl))
        data = DataEntry(comment=cline, data=data)

        datas = (data,)
        for line in tqdm.tqdm(
            fl,
            disable=not progress,
            desc=f"Reading {Path(fname).name}",
            unit="lines",
            leave=leave_progress,
        ):
            cline = CommentLine.read(line, fastspec=fastspec)
            try:
                data = DataLine.read(next(fl))
            except StopIteration:
                # We reached the end of the file.
                break

            try:
                data = DataEntry(comment=cline, data=data)
            except ACQLineError as e:
                warnings.warn(str(e))
                datas = ()
                continue

            if data.comment.swpos == len(datas):
                # Add this to the cycle
                datas += (data,)
            elif data.comment.swpos == 0:
                # Discard the previous cycle and start again -- it was incomplete.
                datas = (data,)
            else:
                # Discard everything and keep going.
                datas = ()
                continue

            if data.comment.swpos == 2:
                # We have a full cycle
                p0.append(datas[0].data.spectrum)
                p1.append(datas[1].data.spectrum)
                p2.append(datas[2].data.spectrum)
                anc.append(datas)
                datas = ()

    anc.complete()
    p0 = np.array(p0)
    p1 = np.array(p1)
    p2 = np.array(p2)

    # Get Q ratio -- need to get this to have compatibility of output with new
    # fastspec default output (HDF5).
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        Q = (p0 - p1) / (p2 - p1)

    if meta:
        return Q.T, [p0.T, p1.T, p2.T], anc
    else:
        return Q, [p0.T, p1.T, p2.T]


def convert_file(
    fname: str | Path,
    outfile: str | None | Path = None,
    write_format: str = "h5",
    **kwargs,
):
    """Convert an ACQ file to a different format.

    Parameters
    ----------
    fname
        The filename of the ACQ file to convert.
    outfile
        The path to the output file. If None, save to a file of the same name (with
        different extension) as ``fname``.
    write_format
        The format to write to. Options are 'h5', 'mat' and 'npz'.

    Other Parameters
    ----------------
    All other parameters passed through to :func:`decode_file`.
    """
    kwargs["meta"] = True
    if write_format not in writers._WRITERS:
        raise ValueError(
            f"The format '{write_format}' does not have an associated writer."
        )

    Q, p, anc = decode_file(fname, **kwargs)

    getattr(writers, f"_write_{write_format}")(
        outfile=outfile or fname,
        ancillary=anc.meta,
        Qratio=Q,
        time_data=anc.data,
        freqs=anc.frequencies,
        fastspec_version=anc.fastspec_version,
        size=len(anc.data["adcmax"]),
        **{f"p{i}": p[i] for i in range(3)},
    )
    return Q, p, anc


def encode(
    filename: str | Path,
    p: list[np.ndarray],
    meta: dict,
    ancillary: dict[str, np.ndarray],
):
    """
    Encode raw powers and ancillary data as an ACQ file.

    Parameters
    ----------
    filename : path
        Path to output file to write.
    p : list of ndarray
        List of three ndarrays, one for each switch. Each array should be 2D, shape
        Ntimes x Nfreq.
    meta : dict
        Dictionary of metadata associated with the file, to be written to the header
        and preambles.
    ancillary : structured array
        Time-dependent ancillary information, such as times and adcmin/adcmax.
    """
    p = np.array(p)

    # data_drops is optional in the .h5 file because we don't really need it.
    data_drops = ancillary.get("data_drops", 0)

    time_has_3dim = len(ancillary["times"].shape) == 2

    with open(filename, "w") as fl:
        # Write the header
        for k, v in meta.items():
            fl.write(f";--{k}: {v}\n")

        # Go through each time
        for i in range(len(p[0])):
            for switch, pp in enumerate(p[:, i]):
                dd = (
                    data_drops[i, switch]
                    if hasattr(data_drops, "__len__")
                    else data_drops
                )
                fl.write(
                    f"# swpos {switch} "
                    f"data_drops {dd:>4} "
                    f"adcmax  {ancillary['adcmax'][i, switch]} "
                    f"adcmin {ancillary['adcmin'][i, switch]} "
                    f"temp  {meta['temp']} C "
                    f"nblk {meta['nblk']} "
                    f"nspec {meta['nfreq']}\n"
                )

                time = ancillary["times"][i]
                if time_has_3dim:
                    time = time[switch]

                fl.write(
                    f"{time} {switch} {meta['freq_min']}  "
                    f"{meta['freq_res']}  {meta['freq_max']}  "
                    f"0.3 spectrum "
                )

                fl.write(_encode_line(pp, meta["nblk"]))
                fl.write("\n")


def convert_h5(infile, outfile):
    """Convert a HDF5 file into an ACQ file.

    HDF5 file must be in the new format used by fastspec.
    """
    try:
        from edges_io.h5 import HDF5RawSpectrum
    except ImportError:  # pragma: no cover
        raise ImportError(
            "To use convert_h5, you need to install edges_io or do "
            "`pip install read_acq[h5]`"
        )

    obj = HDF5RawSpectrum(infile)

    p = [obj["spectra"]["p0"], obj["spectra"]["p1"], obj["spectra"]["p2"]]
    meta = obj["meta"]

    ancillary = {}
    ancillary["adcmax"] = obj["time_ancillary"]["adcmax"]
    ancillary["adcmin"] = obj["time_ancillary"]["adcmin"]
    ancillary["times"] = obj["time_ancillary"]["times"]

    encode(outfile, p=[pp.T for pp in p], meta=meta, ancillary=ancillary)



import pytest

import h5py
import numpy as np
from pathlib import Path
from scipy.io import loadmat

from read_acq import convert_file


@pytest.fixture(scope="module")
def sample_acq() -> Path:
    return Path(__file__).parent / "data/sample.acq"


@pytest.fixture(scope="module")
def tmpdir(tmp_path_factory):
    return tmp_path_factory.mktemp("direc")


def test_h5(tmpdir, sample_acq):
    outfile = tmpdir / "tempfile.h5"
    Q, p, meta = convert_file(sample_acq, outfile=outfile, write_format="h5", meta=True)

    with h5py.File(outfile, "r") as fl:
        qq = fl["spectra"]["Q"][...]

    assert np.allclose(qq[~np.isnan(qq)], Q[~np.isnan(Q)])


def test_mat(tmpdir, sample_acq):
    outfile = tmpdir / "tempfile.mat"
    Q, p, meta = convert_file(
        sample_acq, outfile=outfile, write_format="mat", meta=True
    )

    matdata = loadmat(str(outfile))

    assert np.allclose(matdata["Qratio"][~np.isnan(Q)], Q[~np.isnan(Q)])


def test_npz(tmpdir, sample_acq):
    outfile = tmpdir / "tempfile"
    Q, p, meta = convert_file(
        sample_acq, outfile=outfile, write_format="npz", meta=True
    )

    matdata = np.load(outfile.with_suffix(".npz"))

    assert np.allclose(matdata["Qratio"][~np.isnan(Q)], Q[~np.isnan(Q)])


def test_bad_writer(tmpdir, sample_acq):
    outfile = tmpdir / "tempfile"
    with pytest.raises(ValueError):
        convert_file(sample_acq, outfile=outfile, write_format="bad_format", meta=True)
